/* Text Printer class implementation *\
\* By Justin Pease                   */

#include "gamex.hpp"

TextPrinter tp;

// TextPrinter - Initialize
// Load text font image
void TextPrinter::Initialize(ImageX * alphabetImage, int charsAcrossImage, const char * charListInImageOrder)
{
	assert(alphabetImage != NULL);
	assert(charsAcrossImage >= 1);

	charPixels = alphabetImage->GetWidth() / charsAcrossImage;
	fCharPixels = (float)alphabetImage->GetWidth() / (float)charsAcrossImage;

	assert(charPixels > 3); // make sure the letters aren't ridiculously tiny

	text_img = alphabetImage;

	int numChars = strlen(charListInImageOrder);

	int totalTop = 0;
	int totalBot = 0;
	int total = 0;

	// construct lookup info for each character
	for(int i = 0 ; i < 256 ; i++)
	{
		char c = (char)i;
		int xs = -1, ys = -1;

		chars[i].visible = false;
		for(int j = 0 ; j < numChars ; j++)
		{
			if(charListInImageOrder[j] == c) // if it's the current character
			{
				xs = j % charsAcrossImage;
				ys = j / charsAcrossImage;
				chars[i].visible = true;
				break;
			}
		}

///		// c->(x,y) part of lookup table
///		GetCharOffsets(c, xs, ys);

		chars[i].xs = xs*charPixels;
		chars[i].ys = ys*charPixels;

		// initialize min and max ranges to max and min possible, respectively
		for(int interval = 0 ; interval < HEIGHT_LEVELS ; interval++)
		{
			chars[i].xl[interval] = 255;
			chars[i].xr[interval] = 0;
		}

		if(chars[i].visible)
		{
			// now for the hard part: sample the image for the character's horizontal extents at each height range

			int r,g,b;
			int minX = chars[i].xs, maxX = chars[i].xs + charPixels;
			int minY = chars[i].ys, maxY = chars[i].ys + charPixels;

			if(minX < 0 || maxX < 0 || minY < 0 || maxY < 0 || minX > text_img->GetWidth() || maxX > text_img->GetWidth() || minY > text_img->GetHeight() || maxY > text_img->GetHeight())
			{
				chars[i].visible = false;
				continue;
			}

			assert(minX >= 0 && minY >= 0 && maxX <= text_img->GetWidth() && maxY <= text_img->GetHeight());
			bool findHeights = (c >= 'A' && c <= 'Z');
			bool foundTop = false;
			int botSoFar = 0;

			GameX.AccessPixels(text_img);

			for(int y = minY ; y < maxY ; y++)
			{
				int interval = HEIGHT_LEVELS * (y-minY) / (maxY-minY);
				assert(interval >= 0 && interval < HEIGHT_LEVELS);
				bool foundSomething = false;

				// go from left to right, set min when we find something
				for(int x = minX ; x < maxX ; x++)
				{
					GameX.ReadPixel(x,y, r,g,b);
					b = max(r,max(g,b));

					if(b > GameX.win_maskming)
					{
						if(!foundTop && findHeights)
						{
							totalTop += y-minY;
							total++;
							foundTop = true;
						}
						botSoFar = max(botSoFar, y);

						int pos = (x - minX) * 256 / charPixels;
						assert(pos >= 0 && pos < 256);

						if(chars[i].xl[interval] > pos)
							chars[i].xl[interval] = pos;
						foundSomething = true;
						break;
					}
				}

				// go from right to left, set max when we find something
				if(foundSomething)
				for(int x = maxX-1 ; x >= minX ; x--)
				{
					GameX.ReadPixel(x,y, r,g,b);
					b = max(r,max(g,b));

					if(b > GameX.win_maskming)
					{
						int pos = (x - minX) * 256 / charPixels;
						assert(pos >= 0 && pos < 256);

						if(chars[i].xr[interval] < pos)
							chars[i].xr[interval] = pos;
						break;
					}
				}

				if(chars[i].xr[interval] < 255)
					chars[i].xr[interval]++; // add tiny bit of spacing
			}

			if(foundTop)
				totalBot += botSoFar-minY;

			GameX.EndPixelAccess();
		}

		// special case the white space since there's no image data to go by for it
		for(int interval = 0 ; interval < HEIGHT_LEVELS ; interval++)
		{
			chars[(unsigned char)' '].xl[interval] = 0;
			chars[(unsigned char)' '].xr[interval] = 64;

			chars[(unsigned char)'\t'].xl[interval] = 0;
			chars[(unsigned char)'\t'].xr[interval] = 255;

			chars[0].xl[interval] = 0;
			chars[0].xr[interval] = 255;
		}

	}

	_avgTop = ((total) ? (float)totalTop / (float)total - 2 : 0) / fCharPixels;
	_avgBot = ((total) ? (float)totalBot / (float)total + 2 : fCharPixels) / fCharPixels;
}

void TextPrinter::Initialize(ImageX * alphabetImage, int charsAcrossImage)
{
	Initialize(alphabetImage,charsAcrossImage,"~!@#$%^&*()_+,./`0123456789-=;:?ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz<>\\|[]{}");
}

void TextPrinter::PrintTextScreenProportional(const char * str, float x, float y, float heightRatio, float xStretch, bool centered, int draw_mode, ColorX & color, float angle, bool rotate_baseline, bool fixed)
{
	if(xStretch <= 0.0f || heightRatio <= 0.0f) return;
	float font_size = (heightRatio * GameX.GetHeight()) / (_avgBot - _avgTop);
	float drawYTop = y - (font_size * _avgTop);
	float scaley = 1.0f / xStretch;
	PrintText(str, x, drawYTop, font_size * xStretch, centered, draw_mode, color, angle, scaley, rotate_baseline, fixed);
}

float TextPrinter::GetVisibleStringWidthProportional(const char * str, float heightRatio, float xStretch)
{
	if(xStretch <= 0.0f || heightRatio <= 0.0f) return 0;
	float font_size = (heightRatio * GameX.GetHeight()) / (_avgBot - _avgTop);
	return GetVisibleStringWidth(str,0,32767,font_size * xStretch);
}

float TextPrinter::GetVisibleStringWidthProportional(const char * str, int first_index, int last_index, float heightRatio, float xStretch)
{
	if(xStretch <= 0.0f || heightRatio <= 0.0f) return 0;
	float font_size = (heightRatio * GameX.GetHeight()) / (_avgBot - _avgTop);
	return GetVisibleStringWidth(str,first_index,last_index,font_size * xStretch);
}

// Prints out a string of text graphically using text_img which was loaded from bmp
void TextPrinter::PrintText (const char * str, float x, float y, float font_size, bool centered, int draw_mode, ColorX & color, float angle, float scaley, bool rotate_baseline, bool fixed)
{
	if(font_size <= 0.0f) return;

	if(text_img == NULL)
	{
		int r=0, g=0, b=0;
		if((GameX.GetDrawMode() & DRAW_SUBTRACT) == 0)
			r = color.GetRed(), g = color.GetGreen(), b = color.GetBlue();
		GameX.DrawGDIText((int)x,(int)y,str,r,g,b);

		static bool errAlready = false;
		if(!errAlready)
		{
			GameX.WarningDialog("PrintText being called on TextPrinter that has not been initialized with a valid ImageX.\n(Try putting \"text.gif\" into your game's directory.)\nSubstituting GDI font...");
			errAlready = true;
		}
		return;
	}

	float cosAngle = cosf(angle*DEGtoRAD);
	float sinAngle = sinf(angle*DEGtoRAD);

	if(angle == 0.0f)
		rotate_baseline = false;

	if(centered) {
		float xp = -GetVisibleStringWidth(str,font_size)/2.0f;
		float yp = -font_size/2.0f;
		if(/*angle!=0.0f &&*/ rotate_baseline) {
			x += xp * cosAngle + yp * sinAngle;
			y += xp * sinAngle + yp * cosAngle;
		} else {
			x += xp;
			y += yp;
		}
	}
/*
	if(draw_mode & DRAWOP_CENTERED) {
		float xp = font_size/2.0f - GetVisibleStringWidth(str,font_size)/2.0f;
		if(angle!=0.0f && rotate_baseline) {
			x += xp * cosAngle;
			y += xp * sinAngle;
		} else {
			x += xp;
		}
	}
*/
	inHighlight = false;

	float xi = x;
	float yi = y;
	float scale = font_size / fCharPixels;
	scaley *= scale;
	int n = 0;
	while(str[n]) {
		int xs, ys; // x and y of source position (top-left)
		float kern;
		bool visible; // becomes false if this character is invisible or unknown and thus we don't have to draw it

		GetVisibleCharInfo(str,n,xs,ys,kern,visible);

		if(str[n] == highlightEndChar)
			inHighlight = false;

		// offset letter g down a bit, since unfortunately it's based lower than all other characters
///		if(str[n] == 'g')
///			y += font_size/16.0f;

		float xd, yd;
		if(rotate_baseline) {
			x -= xi;
			y -= yi;
			xd = x * cosAngle - y * sinAngle;
			yd = x * sinAngle + y * cosAngle;
			xd += xi;
			yd += yi;
			x += xi;
			y += yi;
		} else {
			xd = x;
			yd = y;
		}

		if(visible) {
			GameX.SetDrawAngle(-angle);
			GameX.SetDrawScale(scale, scaley);
			GameX.SetDrawShading(inHighlight ? highlightColor : color);
			GameX.SetDrawPart(xs,ys,xs+charPixels,ys+charPixels);
			GameX.SetDrawMode(draw_mode);
			GameX.DrawImage(text_img, (int)xd, (int)yd);
		}

		if(fixed)
			x += font_size;
		else
			x += font_size*kern;

///		if(str[n] == 'g')
///			y -= font_size/16.0f; // reverse earlier vertical offset for letter g

		if(str[n] == highlightStartChar)
			inHighlight = true;

		n++;
	}
}

float TextPrinter::GetVisibleCharWidth(const char * str, int index, float font_size)
{
	return GetVisibleStringWidth(str,index,index,font_size);
}

float TextPrinter::GetVisibleStringWidth(const char * str, float font_size)
{
	return GetVisibleStringWidth(str,0,32767,font_size);
}

float TextPrinter::GetVisibleStringWidth(const char * str, int first_index, int last_index, float font_size)
{
	if(font_size <= 0.0f) return 0.0f;
	int xs,ys; float kern; bool visible;
	float width = 0.0f;
	for(int n = first_index ; n <= last_index && str[n] ; n++) {
		GetVisibleCharInfo(str,n,xs,ys,kern,visible);
		width += kern*font_size;
	}
	return width;
}

void TextPrinter::GetVisibleCharInfo(const char * str, int n, int & xs, int & ys, float & kern, bool & visible)
{
	unsigned char c = str[n], nc = str[n+1];
	xs = chars[c].xs;
	ys = chars[c].ys;
	visible = chars[c].visible;

	assert(xs >= 0 && ys >= 0 || !visible);

	unsigned char minPos = 0;
	for(int i = 0 ; i < HEIGHT_LEVELS ; i++)
	{
		if(chars[nc].xl[i]+minPos < chars[c].xr[i])
			minPos = chars[c].xr[i];
	}

	kern = (float)minPos / 256.0f;
}

// DELETEME (unused)
void TextPrinter::GetCharOffsets(char c, int & xs, int & ys) {}
